﻿abstract class Person {
	public required string Name { get; init; }
	public required string Password { get; init; }

}

class Student : Person {
	public required DateTime DateEnrolled { get; init; }
}

class Teacher : Person {
	public required int CoursesHeld { get; init; }
}

//

abstract class PersonBuilder {
	protected string _name = "";
	protected string _password = "";

	public PersonBuilder WithName(string name) {
		_name = name;
		return this;
	}

	public PersonBuilder WithPassword(string password) {
		_password = password;
		return this;
	}
}

class StudentBuilder : PersonBuilder {
	private DateTime _dateEnrolled;

	public StudentBuilder WithDateEnrolled(DateTime dateEnrolled) {
		_dateEnrolled = dateEnrolled;
		return this;
	}

	public Student Build() => new Student { DateEnrolled = _dateEnrolled, Name = _name, Password = _password };
}

class TeacherBuilder : PersonBuilder {
	private int _coursesHeld;

	public TeacherBuilder WithCoursesHeld(int coursesHeld) {
		_coursesHeld = coursesHeld;
		return this;
	}

	public Teacher Build() => new Teacher { CoursesHeld =  _coursesHeld, Name = _name, Password = _password };
}

class Program {
	static void Main(string[] args) {
		var studentA = new StudentBuilder()
							.WithName("Pavel Rozek").WithPassword("lavice")
							// Error: 'PersonBuilder' does not contain a definition for 'WithDateEnrolled' and no accessible extension method 'WithDateEnrolled' accepting a first argument of type 'PersonBuilder' could be found (are you missing a using directive or an assembly reference ?)
							.WithDateEnrolled(new DateTime(2015, 9, 21));
		var studentB = new StudentBuilder()
							.WithName("Marie Pichova").WithPassword("zidle")
							// Error: 'PersonBuilder' does not contain a definition for 'WithDateEnrolled' and no accessible extension method 'WithDateEnrolled' accepting a first argument of type 'PersonBuilder' could be found (are you missing a using directive or an assembly reference ?)
							.WithDateEnrolled(new DateTime(2018, 9, 25));
		var teacher = new TeacherBuilder()
							.WithName("Pavel Jezek").WithPassword("stul")
							// Error: 'PersonBuilder' does not contain a definition for 'WithCoursesHeld' and no accessible extension method 'WithCoursesHeld' accepting a first argument of type 'PersonBuilder' could be found (are you missing a using directive or an assembly reference ?)
							.WithCoursesHeld(5);
	}
}
