﻿abstract class Person {
	public required string Name { get; init; }
	public required string Password { get; init; }

}

class Student : Person {
	public required DateTime DateEnrolled { get; init; }
}

class Teacher : Person {
	public required int CoursesHeld { get; init; }
}

//

abstract class PersonBuilder<TSelf> where TSelf : PersonBuilder<TSelf> {
	protected string _name = "";
	protected string _password = "";

	public TSelf WithName(string name) {
		_name = name;
		return (TSelf) this;    // Without cast = Error: Cannot implicitly convert type 'PersonBuilder<TSelf>' to 'TSelf'. An explicit conversion exists(are you missing a cast?)

	}
	public TSelf WithPassword(string password) {
		_password = password;
		return (TSelf) this;        // Without cast = Error: Cannot implicitly convert type 'PersonBuilder<TSelf>' to 'TSelf'. An explicit conversion exists(are you missing a cast?)
	}
}

class StudentBuilder : PersonBuilder<StudentBuilder> {
	private DateTime _dateEnrolled;
	public StudentBuilder WithDateEnrolled(DateTime dateEnrolled) {
		_dateEnrolled = dateEnrolled;
		return this;
	}
	public Student Build() => new Student { DateEnrolled = _dateEnrolled, Name = _name, Password = _password };
}

class TeacherBuilder : PersonBuilder<TeacherBuilder> {
	private int _coursesHeld;

	public TeacherBuilder WithCoursesHeld(int coursesHeld) {
		_coursesHeld = coursesHeld;
		return this;
	}

	public Teacher Build() => new Teacher { CoursesHeld = _coursesHeld, Name = _name, Password = _password };
}

class Program {
	static void Main(string[] args) {
		var studentA = new StudentBuilder()
							.WithName("Pavel Rozek").WithPassword("lavice")
							.WithDateEnrolled(new DateTime(2015, 9, 21))
							.Build();
		var studentB = new StudentBuilder()
							.WithName("Marie Pichova").WithPassword("zidle")
							.WithDateEnrolled(new DateTime(2018, 9, 25))
							.Build();
		var teacher = new TeacherBuilder()
							.WithName("Pavel Jezek").WithPassword("stul")
							.WithCoursesHeld(5)
							.Build();
		var teacher2 = new TeacherBuilder()
							.WithPassword("skrin")
							.WithCoursesHeld(5)
							.WithName("Petr Hnetynka")
							.Build();

		List<Person> people = new List<Person> { studentA, studentB, teacher, teacher2 };
		Console.WriteLine("+++ People:");
		people.PrintAll();
	}
}

static class ListPersonExtensions {
	public static void PrintAll(this List<Person> people) {
		foreach (var person in people) {
			Console.WriteLine($"{person.GetType().Name} {person.Name} has password \"{person.Password}\"");
		}
	}
}
