﻿
abstract class Person<TConcretePerson> where TConcretePerson : Person<TConcretePerson>, new() {
	public string Name { get; init; }
	public string Password { get; init; }

	public TConcretePerson WithName(string newName) {
		var copy = CloneConcretePerson();
		copy.Name = newName;
	}

	public TConcretePerson WithPassword(string newPassword) {
		var copy = CloneConcretePerson();
		copy.Password = newPassword;
	}

	protected Person(string name, string password) {
		Name = name;
		Password = password;
	}

	protected abstract TConcretePerson CloneConcretePerson();

	public static TConcretePerson Default { get; } = new TConcretePerson();
}

class Student : Person<Student> {
	public DateTime DateEnrolled { get; init; }

	public Student WithDateEnrolled(DateTime newDateEnrolled) => new Student(Name, Password, newDateEnrolled);

	public Student(string name, string password, DateTime dateEnrolled) : base(name, password) =>
		DateEnrolled = dateEnrolled.Date;

	public Student() : this("", "", default) { }

	protected override Student CloneConcretePerson() =>
		new Student { Name = this.Name, Password = this.Password, DateEnrolled = this.DateEnrolled };
}

class Teacher : Person<Teacher> {
	public int CoursesHeld { get; init; }

	public Teacher WithCoursesHeld(int newCoursesHeld) => new Teacher(Name, Password, newCoursesHeld);

	public Teacher(string name, string password, int coursesHeld) : base(name, password)
		=> CoursesHeld = coursesHeld;

	public Teacher() : this("", "", 0) { }

	protected override Teacher CloneConcretePerson() =>
		new Teacher { Name = this.Name, Password = this.Password, CoursesHeld = this.CoursesHeld };
}

class Program {
	static void Main(string[] args) {
		var studentA = Student.Default
						.WithName("Pavel Rozek").WithPassword("lavice")
						.WithDateEnrolled(new DateTime(2015, 9, 21));

		var studentB = Student.Default
						.WithName("Marie Pichova").WithPassword("zidle")
						.WithDateEnrolled(new DateTime(2018, 9, 25));

		var teacher = Teacher.Default
						.WithName("Pavel Jezek").WithPassword("stul")
						.WithCoursesHeld(5);

		var teacher2 = Teacher.Default
							.WithPassword("skrin")
							.WithCoursesHeld(5)
							.WithName("Petr Hnetynka");
	}
}


