﻿
abstract class Person<TConcretePerson> where TConcretePerson : Person<TConcretePerson>, new() {
	public string Name { get; }
	public string Password { get; }

	public TConcretePerson WithName(string newName) => CloneConcretePersonWith(newName, Password);
	public TConcretePerson WithPassword(string newPassword) => CloneConcretePersonWith(Name, newPassword);

	protected Person(string name, string password) {
		Name = name;
		Password = password;
	}

	protected abstract TConcretePerson CloneConcretePersonWith(string name, string password);

	public static TConcretePerson Default { get; } = new TConcretePerson();
}

class Student : Person<Student> {
	public DateTime DateEnrolled { get; }

	public Student WithDateEnrolled(DateTime newDateEnrolled) => new Student(Name, Password, newDateEnrolled);

	public Student(string name, string password, DateTime dateEnrolled) : base(name, password) =>
		DateEnrolled = dateEnrolled.Date;

	public Student() : this("", "", default) { }

	protected override Student CloneConcretePersonWith(string name, string password) =>
		new Student(name, password, this.DateEnrolled);
}

class Teacher : Person<Teacher> {
	public int CoursesHeld { get; }

	public Teacher WithCoursesHeld(int newCoursesHeld) => new Teacher(Name, Password, newCoursesHeld);

	public Teacher(string name, string password, int coursesHeld) : base(name, password)
		=> CoursesHeld = coursesHeld;

	public Teacher() : this("", "", 0) { }

	protected override Teacher CloneConcretePersonWith(string name, string password) =>
		new Teacher(name, password, this.CoursesHeld);
}

class Program {
	static void Main(string[] args) {
		var studentA = Student.Default
						.WithName("Pavel Rozek").WithPassword("lavice")
						.WithDateEnrolled(new DateTime(2015, 9, 21));

		var studentB = Student.Default
						.WithName("Marie Pichova").WithPassword("zidle")
						.WithDateEnrolled(new DateTime(2018, 9, 25));

		var teacher = Teacher.Default
						.WithName("Pavel Jezek").WithPassword("stul")
						.WithCoursesHeld(5);

		var teacher2 = Teacher.Default
							.WithPassword("skrin")
							.WithCoursesHeld(5)
							.WithName("Petr Hnetynka");

		// Error: Using the generic type 'Person<TConcretePerson>' requires 1 type argument
		// List<Person> people = new List<Person> { studentA, studentB, teacher, teacher2 };
	}
}


